#include <math.h>
#include <stdio.h>
#include <inttypes.h>

#define PI atan2 (0.0, -1.0)
#define DEGREES		256

const double d = 4.0;

uint8_t slope_tab[65536] = { 0 };
int tab_len = 0;
int indexes[65536];
int num_indexes = 0;
int slopes_used[DEGREES] = { 0 };

int lengths[DEGREES] = { 0 };

#include "amiga.h"

//generate slopes
void add_slopes() {
    double st;
    int x, y;
    int i, j;
    uint8_t slope[128];

    int deg;

    tab_len = 0;

    //try to generate a slope per degree
    for (deg = 0; deg < DEGREES; deg++) {
        if (!slopes_used[deg]) fprintf(stderr, "slope %d not used\n", deg);
        st = tan(deg * PI / DEGREES);

        y = 0;
        x = 0;

        //calc slope
        while (y < lengths[deg]) {
            x = st * y;
            slope[y] = x;
            y++;
        }
        //find possible match in already generated data
        for (i = 0; i <= tab_len - y; i++) {
            j = 0;
            while (slope[j] == slope_tab[i+j]) {
                j++;
                if (j == y) {
                    fprintf(stderr, "slope %d matches @$%04x len: %d\n", deg, i, y);
                    indexes[num_indexes] = i;
                    num_indexes++;
                    goto done;
                }
            }
        }
        done:
        //add slope if no match occured
        if (j != y) {
            indexes[num_indexes++] = tab_len;
            if(lengths[deg]) {
                for (i = 0; i < y; i++) {
                    slope_tab[tab_len + i] = slope[i];
                }
                tab_len += i;
            }
        }
    }
    fprintf(stderr, "costs for slopes: $%04x (+$0100 index tabs)\n", tab_len);
}

int slope(int x1, int y1, int x2, int y2) {
    double dx, dy;
    int st;
    dy = y2 - y1;
    dx = x2 - x1;

    //if (dy < 0) { dy = -dy; dx = -dx; }

    if (dy == 0) return -1;
    else st = atan(dx/dy) * DEGREES / PI;
    if (st < 0) st += DEGREES;
    return st;
}

int one_slope(int sx1, int sy1, int sx2, int sy2, int sx3, int sy3) {
    double dx1, dy1;
    double dx2, dy2;
    int st1, st2;
    int slope[128];

    int i;

    dy1 = sy2 - sy1;
    dx1 = sx2 - sx1;
    dy2 = sy3 - sy2;
    dx2 = sx3 - sx2;

    if (dy1 == 0 || dy2 == 0) return -1;

    st1 = dx1/dy1;
    st2 = dx2/dy2;

    if (dy1 < 0) { dy1 = -dy1; dx1 = -dx1; }
    if (dy2 < 0) { dy2 = -dy2; dx2 = -dx2; }

    for (i = 0; i <= dy1 + dy2; i++) {
        slope[i] = st1 * i;
    }
    for (i = 0; i <= dy2; i++) {
        if (slope[i + (int)dy1] != slope[(int)dy1] + st2 * i) break;
    }
    if (i <= dy2) return -1;
    return 1;
}

void add_face(int* rx, int* ry, int* lx, int* ly) {
    int l,r;
    double dx;
    double dy;
    int st;

    int refill = 0;
    int old_refill = 0;

    int sl_l;
    int sl_r;

    int y_max;

    l = 0;
    r = 0;

    y_max = 0;
    for (st = 0; st < 4; st++) if (ly[st] > y_max) y_max = ly[st];

    while (r + l < 4) {
        if (r == 0) {
            while ((dy = ry[(r+1) & 3] - ry[r]) == 0) r++;
            dx = rx[(r+1) & 3] - rx[r];

            if (dy == 0) st = 0;
            else st = atan(dx/dy) * DEGREES / PI;
            if (st < 0) st += DEGREES;
            if (lengths[st] < dy) lengths[st] = dy;
            sl_r = st;
            slopes_used[st]++;

            while ((dy = ly[(l+1) & 3] - ly[l]) == 0) l++;
            dx = lx[(l+1) & 3] - lx[l];

            if (dy == 0) st = 0;
            else st = atan(dx/dy) * DEGREES / PI;
            if (st < 0) st += DEGREES;
            //remember longest slope
            if (lengths[st] < dy) lengths[st] = dy;
            sl_l = st;
            slopes_used[st]++;

            //evaluate next refill position
            if (ly[l+1] < ry[r+1]) refill = ly[l+1];
            else refill = ry[r+1];

            printf("!byte $%02x ;y_start | $80\n",ly[l] | 0x80);
            if (lx[l] == rx[r]) {
                printf("!byte $%02x ;-lrx_start & $7f\n",(-lx[l]) & 0x7f);
            } else {
                printf("!byte $%02x ;-lrx_start & $7f\n",((-lx[l]) & 0x7f) | 0x80);
                printf("!byte $%02x ;-rx_start & $7f\n",(-rx[r]) & 0x7f);
            }
            printf("!byte $%02x ;r_slope[0]\n",sl_r);
            printf("!byte $%02x ;l_slope[0]\n",sl_l);
            //if (sl_l != 0 && sl_r != 0) printf("!byte $%02x\n",refill);
            printf("!byte $%02x\n",refill);

            l++;
            r++;

            old_refill = refill;

        } else {
            //skip identical plots
            if (ly[l] == ly[(l+1)&3] && lx[l] == lx[(l+1)&3]) l++;
            if (ry[r] == ry[(r+1)&3] && rx[r] == rx[(r+1)&3]) r++;

            //refill l
            if (ly[l&3] <= ry[r&3]) {
                //evaluate next refill position
                if (r + l < 4) {
                    if (ly[(l+1)&3] < ry[r&3]) {
                        refill = ly[(l+1)&3];
                    } else {
                        refill = ry[r&3];
                    }
                    if (old_refill == y_max) return;
                    old_refill = refill;
                    printf("!byte $%02x\n",refill);
                    dx = lx[(l+1) & 3] - lx[l&3];
                    dy = ly[(l+1) & 3] - ly[l&3];

                    if (dy == 0) st = 0;
                    else st = atan(dx/dy) * DEGREES / PI;
                    if (st < 0) st += DEGREES;
                    //remember longest slope
                    if (lengths[st] < dy) lengths[st] = dy;
                    printf("!byte $%02x, $%02x\n", (-lx[l&3]) & 0xff, st);
                    l++;
                    slopes_used[st]++;
                }
            }
            //refill r
            if (ly[l&3] >= ry[r&3]) {
                //evaluate next refill position
                if (r + l < 4) {
                    if (ly[l&3] < ry[(r+1)&3]) {
                        refill = ly[l&3];
                    } else {
                        refill = ry[(r+1)&3];
                    }
                    if (old_refill == y_max) return;
                    old_refill = refill;
                    printf("!byte $%02x\n",refill);
                    dx = rx[(r+1) & 3] - rx[r&3];
                    dy = ry[(r+1) & 3] - ry[r&3];

                    if (dy == 0) st = 0;
                    else st = atan(dx/dy) * DEGREES / PI;
                    if (st < 0) st += DEGREES;
                    //remember longest slope
                    if (lengths[st] < dy) lengths[st] = dy;
                    printf("!byte $%02x, $%02x\n", (-rx[r&3]) & 0x7f, st);
                    r++;
                    slopes_used[st]++;
                }
            }
        }
    }
}

void calc_slope(int* slope, int x1, int y1, int x2, int y2) {
    double dx = x2 - x1;
    double dy = y2 - y1;
    double st = dx/dy;

    int y;
    for (y = 0; y <= dy; y++) {
        slope[y1 + y] = x1 + (st * y);
    }
}

int test_corridor(int* tx, int* ty, int* sx, int* sy) {
    int i, j;
    int y;

    int sslope[128] = {0};
    int tslope[128] = {0};

    for (i = 0; i < 4; i++) {
        //y_max reached?!
        if (sy[i] > sy[(i+1) & 3]) break;
        //not yet, continue with calculating slope
        calc_slope(sslope, sx[i], sy[i], sx[(i+1) & 3], sy[(i+1) & 3]);
    }

    for (j = 0; j < 4; j++) {
        if (ty[j] > ty[(j+1) & 3]) break;
        calc_slope(tslope, tx[j], ty[j], tx[(j+1) & 3], ty[(j+1) & 3]);
    }

    for (y = sy[0]+1; y < sy[i & 3]; y++) {
        //check if any parts of the to be checked slope are on the right side of our slope
        if (tslope[y] !=0 && sslope[y] != 0 && tslope[y] > sslope[y]) return 0;
    }
    //now we have all x valuues for y_min to y_max on right side;
    return 1;
}

int main () {
    int i, frame;
    double deg_x = 0.0;
    double deg_y = 0.0;
    double deg_z = 0.0;

    double t_x[num_vertices];
    double t_y[num_vertices];

    double x, y, z;
    double x_, y_, z_;

    double sam;

    int max_frames = 128;

    int saved;
    int y_min, j, k;
    int ty_min;

    int done[256];

    int processed;
    int old_p;
    int val;

    int tx[4];
    int ty[4];
    int lx[4];
    int ly[4];

    int rx[4];
    int ry[4];

    for (frame = 0; frame < max_frames; frame++) {

        deg_x = 2 * PI / (double)max_frames * frame * 3.0;
        deg_y = 2 * PI / (double)max_frames * frame * 2.0;
        deg_z = 2 * PI / (double)max_frames * frame * 1.0;

        for (i = 0; i < num_vertices; i++) {
            //rotate this shit
            x_ = vertx[i] * cosf(deg_z) - verty[i] * sinf(deg_z);
            y_ = vertx[i] * sinf(deg_z) + verty[i] * cosf(deg_z);

            x = x_;
            y = y_;

            y_ = y * cosf(deg_y) - vertz[i] * sinf(deg_y);
            z_ = y * sinf(deg_y) + vertz[i] * cosf(deg_y);

            y = y_;
            z = z_;

            x_ = z * sinf(deg_x) + x * cosf(deg_x);
            z_ = z * cosf(deg_x) - x * sinf(deg_x);

            //perspective
            t_x[i] = ((x_ / (1.0 + z_ / d)) * scaling + 64.0)+4;
            t_y[i] = ((y_ / (1.0 + z_ / d)) * scaling + 64.0);
        }
        for (i = 0; i < 256; i++) done[i] = 0;

        processed = 0;
        for (i = 0; i < num_faces; i++) {
            //culling via signed area
            sam = (t_y[faces[i][1]] - t_y[faces[i][0]]) * (t_x[faces[i][2]] - t_x[faces[i][1]]) - (t_x[faces[i][1]] - t_x[faces[i][0]]) * (t_y[faces[i][2]] - t_y[faces[i][1]]);
            //mark culled faces as done
            if(sam <= 0) {
                done[i] = 1;
                processed++;
            }
        }
        while (processed < num_faces) {
            old_p = processed;
            for (i = 0; i < num_faces; i++) {
                if (!done[i]) {

                    y_min = 0;
                    for (k = 0; k < 4; k++) {
                        if (t_y[faces[i][y_min]] >= t_y[faces[i][k]] ) y_min = k;
                    }

                    //face to be drawn with left side first
                    lx[0] = t_x[faces[i][(y_min + 0) & 3]];
                    lx[1] = t_x[faces[i][(y_min + 1) & 3]];
                    lx[2] = t_x[faces[i][(y_min + 2) & 3]];
                    lx[3] = t_x[faces[i][(y_min + 3) & 3]];

                    ly[0] = t_y[faces[i][(y_min + 0) & 3]];
                    ly[1] = t_y[faces[i][(y_min + 1) & 3]];
                    ly[2] = t_y[faces[i][(y_min + 2) & 3]];
                    ly[3] = t_y[faces[i][(y_min + 3) & 3]];

                    //check if we can safely draw that face (all stuff in a corridor right from us must have been drawn first)
                    for (j = 0; j < num_faces; j++) {
                        //only check faces that were not processed yet, and don't check against ourself
                        if (!done[j] && j != i) {
                            ty_min = 0;
                            for (k = 0; k < 4; k++) {
                                if (t_y[faces[j][ty_min]] >= t_y[faces[j][k]] ) ty_min = k;
                            }

                            //face to be checked
                            tx[0] = t_x[faces[j][(ty_min + 0) & 3]];
                            tx[1] = t_x[faces[j][(ty_min + 3) & 3]];
                            tx[2] = t_x[faces[j][(ty_min + 2) & 3]];
                            tx[3] = t_x[faces[j][(ty_min + 1) & 3]];

                            ty[0] = t_y[faces[j][(ty_min + 0) & 3]];
                            ty[1] = t_y[faces[j][(ty_min + 3) & 3]];
                            ty[2] = t_y[faces[j][(ty_min + 2) & 3]];
                            ty[3] = t_y[faces[j][(ty_min + 1) & 3]];

                            val = test_corridor(tx, ty, lx, ly);
                            if (val == 0) break;
                        }
                    }
                    //no collision detected, so spit out face
                    if (j == num_faces) {
                        done[i] = 1;
                        processed++;

                        //if same slope follows another, aggregate vertices, to a single slope, but only if none of the slopes has a dy of 0
                        if (slope(lx[0], ly[0], lx[1], ly[1]) == slope(lx[1], ly[1], lx[2], ly[2]) && lx[2] - lx[1] != 0 && ly[2] - ly[1] != 0 && lx[1] - lx[0] != 0 && ly[1] - ly[0] != 0) {
                            lx[0] = lx[1];
                            ly[0] = ly[1];
                        } else if (slope(lx[1], ly[1], lx[2], ly[2]) == slope(lx[2], ly[2], lx[3], ly[3]) && lx[3] - lx[0] != 0 && ly[3] - ly[0] != 0 && lx[2] - lx[1] != 0 && ly[2] - ly[1] != 0) {
                            lx[1] = lx[2];
                            ly[1] = ly[2];
                        } else if (slope(lx[2], ly[2], lx[3], ly[3]) == slope(lx[3], ly[3], lx[0], ly[0]) && lx[0] - lx[3] != 0 && ly[0] - ly[3] != 0 && lx[3] - lx[2] != 0 && ly[3] - ly[2] != 0) {
                            lx[2] = lx[3];
                            ly[2] = ly[3];
                        } else if (slope(lx[3], ly[3], lx[0], ly[0]) == slope(lx[0], ly[0], lx[1], ly[1]) && lx[1] - lx[0] != 0 && ly[1] - ly[0] != 0 && lx[0] - lx[3] != 0 && ly[0] - ly[3] != 0) {
                            lx[3] = lx[0];
                            ly[3] = ly[0];
                        }

                        //face to be drawn with right side first -> change clockwiseness
                        rx[0] = lx[0];
                        ry[0] = ly[0];
                        rx[1] = lx[3];
                        ry[1] = ly[3];
                        rx[2] = lx[2];
                        ry[2] = ly[2];
                        rx[3] = lx[1];
                        ry[3] = ly[1];

                        saved = 0;
                        //check if vertices match others
                        if (rx[0] == rx[3] && ry[0] == ry[3]) saved++;
                        if (rx[3] == rx[2] && ry[3] == ry[2]) saved++;
                        if (rx[2] == rx[1] && ry[2] == ry[1]) saved++;
                        if (rx[1] == rx[0] && ry[1] == ry[0]) saved++;

                        //one or less vertices matching, so we still describe an area and not a line or dot, all fine
                        if (saved <= 1) {
                            saved = 0;
                            //sanity check
                            //dy is 3 times 0? we can't draw that
                            if (ly[0] - ly[1] == 0) saved++;
                            if (ly[1] - ly[2] == 0) saved++;
                            if (ly[2] - ly[3] == 0) saved++;
                            if (ly[3] - ly[0] == 0) saved++;

                            if (saved <= 2) {
                                add_face(rx, ry, lx, ly);
                            }
                        }
                    }
                }
            }
            //avoid endless loops
            if (old_p == processed) break;
        }
        printf("!byte $80\n");
    }
    add_slopes();
    FILE* fw;
    fw = fopen("slopes.bin", "wb");
    fwrite(slope_tab,1,tab_len,fw);
    fclose(fw);
    fw = fopen("slope_indexes.bin", "wb");
    for (i = 0; i < num_indexes; i++) {
        j = (indexes[i] + 0x2800) >>8;
        fwrite(&j,1,1,fw);
    }
    for (i = 0; i < num_indexes; i++) {
        j = (indexes[i] + 0x2800) & 0xff;
        fwrite(&j,1,1,fw);
    }
    fclose(fw);
    return 0;
}
